"""
Constants for the pysomneo library.
"""

SOUND_SOURCE_ALARM = {"wake-up": "wus", "radio": "fmr", "off": "off"}
SOUND_SOURCE_DUSK = {"radio": "fmr", "off": "off", "dusk": "dus"}

FM_PRESETS = ["fm 1", "fm 2", "fm 3", "fm 4", "fm 5"]

SOURCES = {
    "AUX": "aux",
    "FM 1": 1,
    "FM 2": 2,
    "FM 3": 3,
    "FM 4": 4,
    "FM 5": 5,
    "Other": "other",
}

DAYS = {
    0: "tomorrow",
    2: "mon",
    4: "tue",
    8: "wed",
    16: "thu",
    32: "fri",
    64: "sat",
    128: "sun",
}
DAYS_TYPE = {0: "tomorrow", 62: "workdays", 192: "weekend", 254: "daily"}

STATUS = {
    1: "off",
    2: "sunset",
    257: "light-on",
    776: "sunset",
    777: "sunset",
    2321: "snooze",
    2309: "wake-up",
    2817: "on",
}
